/* 
 * Copyright (C) 2002 Laird Breyer
 *  
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * Author:   Laird Breyer <laird@lbreyer.com>
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <signal.h>
#include <unistd.h> 

#include "dbacl.h"

extern options_t u_options;
extern options_t m_options;

extern Regex re[2*MAX_RE];
extern regex_count_t regex_count;
extern regex_count_t antiregex_count;

/* global variables */

MBOX_State mbox;
XML_State xml;

char *textbuf = NULL;
charbuf_len_t textbuf_len = 0;
char *aux_textbuf = NULL;
charbuf_len_t aux_textbuf_len = 0;

#if defined HAVE_MBRTOWC
wchar_t *wc_textbuf = NULL;
charbuf_len_t wc_textbuf_len = 0;
#endif

token_order_t ngram_order = 1; /* default */

int sa_signal = 0;

extern long system_pagesize;
extern void *in_iobuf;
extern void *out_iobuf;

extern int cmd;

/***********************************************************
 * EXPERIMENTAL:                                           *
 * this code is an experiment to see if memory mapping is  *
 * faster than buffered I/O. Surprisingly, the gain is at  *
 * best marginal (less than two percent on my test).       *
 * The code is left in for future experiments              *
 *                                                         *
 * I suspect that mmap() isn't faster because we only read *
 * the input files sequentially, never backtrack, and my   *
 * Debian development system probably reads ahead quite    *
 * agressively. So normal buffered I/O does about the same *
 * amount of work, In any case, most of the time is spent  *
 * in the hash tables and calculations anyway.             *
 *                                                         *
 * The mmap() code below is written so as to easily replace*
 * the buffered I/O functions                              *
 ***********************************************************/

#undef EXPERIMENTAL

#if defined EXPERIMENTAL

#include <sys/stat.h>
#include <errno.h>

typedef struct {
  FILE *f;
  size_t fsize;
  size_t fmappos;
  size_t fmapsize;
  char *seekpos;
  char *mstart;
} MMFILE;

int MMEOF(MMFILE *stream) {
  return stream->mstart && ((stream->seekpos - stream->mstart) + 
			    stream->fmappos >= stream->fsize);
}

MMFILE *MMOPEN(FILE *f) {
  MMFILE *stream = NULL;
  struct stat sb;
  if( f ) {
    stream = (MMFILE *)malloc(sizeof(MMFILE));
    if( stream ) {
      stream->f = f;
#define MEGABYTES 20
      stream->fmapsize = MEGABYTES*1024*1024;
      if( fstat(fileno(f), &sb) == -1 ) {
	free(stream);
	return NULL;
      }
      stream->fsize = sb.st_size;
      stream->fmappos = 0;
      stream->mstart = mmap(NULL, stream->fmapsize, PROT_READ, MAP_SHARED,
			    fileno(stream->f), stream->fmappos);
      if( stream->mstart == MAP_FAILED ) {
	free(stream);
	return NULL;
      }
      stream->seekpos = stream->mstart;
#if defined HAVE_MADVISE
      if( stream->mstart ) {
	madvise(stream->mstart, stream->fmapsize, MADV_SEQUENTIAL);
      }
#endif
    }
  }
  return stream;
}

int MMCLOSE(MMFILE *stream) {
  int r = 0;
  if( stream ) {
    r = munmap(stream->mstart, stream->fmapsize);
    free(stream);
  }
  return r;
}

void MMFORWARD(MMFILE *stream) {
  size_t offset;
  if( stream->fmappos + stream->fmapsize < stream->fsize ) {
    offset = stream->seekpos - stream->mstart;
    if( munmap(stream->mstart, stream->fmapsize) == -1 ) {
      free(stream);
      exit(0);
    }

    stream->fmappos += system_pagesize * (offset/system_pagesize);
    stream->mstart = mmap(0, stream->fmapsize, PROT_READ, MAP_SHARED,
			  fileno(stream->f), stream->fmappos);
    if( stream->mstart == MAP_FAILED ) {
      free(stream);
      exit(0);
    }
    stream->seekpos = stream->mstart + 
      (offset - system_pagesize * (offset/system_pagesize));
#if defined HAVE_MADVISE
    if( stream->mstart ) {
      madvise(stream->mstart, stream->fmapsize, MADV_SEQUENTIAL);
    }
#endif
  }
}

char *MMGETS(char *s, size_t size, MMFILE *stream) {
  char *result = NULL;
  size_t left;

  left = stream->fmapsize - (stream->seekpos - stream->mstart);
  if( left < size ) {
    MMFORWARD(stream);
    left = stream->fmapsize - (stream->seekpos - stream->mstart);
  }

  if( left <= 0 ) {
    return NULL;
  } else if( size <= left + 1) {
    result = memccpy(s, stream->seekpos, '\n', size - 1);
    if( result ) { 
      *result = '\0';
      stream->seekpos += result - s;
      return s;
    }
  } else { 
    result = memccpy(s, stream->seekpos, '\n', left);
    if( result ) {
      *result = '\0';
      stream->seekpos = result;
    } else {
      memcpy(s, stream->seekpos, left);
      s[left] = '\0';
      stream->seekpos += left;
    }
    return s;
  }
  return NULL;
}

#endif

/***********************************************************
 * SIGNAL HANDLING                                         *
 ***********************************************************/

/* sigaction structure used by several functions */
struct sigaction act;

void my_sa_handler(int signum) {
  sa_signal = signum;
}

/* intercepts typical termination signals and tries to do the right thing */
void init_signal_handling() {

  memset(&act, 0, sizeof(act));

  /* set up global sigaction structure */

  act.sa_handler = my_sa_handler;
  sigemptyset(&act.sa_mask);
  sigaddset(&act.sa_mask,SIGHUP);
  sigaddset(&act.sa_mask,SIGINT);
  sigaddset(&act.sa_mask,SIGQUIT);
  sigaddset(&act.sa_mask,SIGTERM);
  sigaddset(&act.sa_mask,SIGPIPE);
  sigaddset(&act.sa_mask,SIGUSR1);
  act.sa_flags = 0;

  sigaction(SIGHUP, &act, NULL);
  sigaction(SIGINT, &act, NULL);
  sigaction(SIGQUIT, &act, NULL);
  sigaction(SIGTERM, &act, NULL);
  sigaction(SIGPIPE, &act, NULL);
  sigaction(SIGUSR1, &act, NULL);
}

void cleanup_signal_handling() {
  /* nothing - this is just to mess with your head ;-) */
}

void process_pending_signal(FILE *input) {
  
  if( sa_signal ) {

    sigprocmask(SIG_BLOCK, &act.sa_mask, NULL);

    switch(sa_signal) {
    case SIGINT:
      fprintf(stderr, 
	      "signal: caught interrupt request, exiting.\n");
      exit(1);
      break;
    case SIGPIPE:
      /* should we terminate, or should we ignore? */
      fprintf(stderr,
	      "error: broken pipe on output, exiting.\n");
      exit(1);
      break;
    case SIGHUP:
    case SIGQUIT:
    case SIGTERM:
      fprintf(stderr, 
	      "signal: caught termination request, ignoring further input\n");
      if( input ) { fclose(input); }
      cmd |= (1<<CMD_QUITNOW);
      break;
    case SIGUSR1:
      if( u_options & (1<<U_OPTION_CLASSIFY) ) {
	fprintf(stderr, 
		"signal: caught SIGUSR1 request, reloading categories asap\n");
	cmd |= (1<<CMD_RELOAD_CATS);
      } else {
	fprintf(stderr, 
		"signal: caught SIGUSR1 request, ignoring\n");
      }
      break;
    default:
      /* nothing */
      break;
    }
    sa_signal = 0;

    sigprocmask(SIG_UNBLOCK, &act.sa_mask, NULL);

  }
}

/***********************************************************
 * MISCELLANEOUS FILE HANDLING                             *
 ***********************************************************/

void init_file_handling() {
  /* preallocate primary text holding buffer */
  textbuf_len = system_pagesize;
  textbuf = malloc(textbuf_len);
  memset(&mbox.boundary, 0, sizeof(mbox.boundary));

#if defined HAVE_MADVISE
  madvise(textbuf, sizeof(char) * textbuf_len, MADV_SEQUENTIAL);
#endif

#if defined HAVE_POSIX_MEMALIGN
  /* buffer must exist until after fclose() if used in setvbuf() */
  if( 0 != posix_memalign(&in_iobuf, system_pagesize, 
			  BUFFER_MAG * system_pagesize) ) {
    in_iobuf = NULL; /* just to be sure */
  }
  /* buffer must exist until after fclose() if used in setvbuf() */
  if( 0 != posix_memalign(&out_iobuf, system_pagesize, 
			  BUFFER_MAG * system_pagesize) ) {
    out_iobuf = NULL; /* just to be sure */
  }
#elif defined HAVE_MEMALIGN
  /* memalign()ed memory can't be reclaimed by free() */
  in_iobuf = (void *)memalign(system_pagesize, BUFFER_MAG * system_pagesize);
  out_iobuf = (void *)memalign(system_pagesize, BUFFER_MAG * system_pagesize);
#elif defined HAVE_VALLOC
  /* valloc()ed memory can't be reclaimed by free() */
  in_iobuf = (void *)valloc(BUFFER_MAG * system_pagesize);
  out_iobuf = (void *)valloc(BUFFER_MAG * system_pagesize);
#endif
}

void cleanup_file_handling() {
  /* free some global resources */
  free(textbuf);
  /* no need to reserve space for both char and wide char caches */
  if( m_options & (1<<M_OPTION_I18N) ) {
#if defined HAVE_MBRTOWC
    w_free_decoding_caches();
#endif
  } else {
    free_decoding_caches();
  }
#if defined HAVE_POSIX_MEMALIGN
  if( in_iobuf ) { free(in_iobuf); }
  if( out_iobuf ) { free(out_iobuf); }
#endif
}

/* Given a mime type, this selects an appropriate
 * xml default setting for the character filter.
*/
XML_Reset select_xml_defaults(MIME_Struct *mime) {
  switch(mime->type) {
  case ctUNDEF:
  case ctPLAINTEXT:
    return (m_options & (1<<M_OPTION_NOPLAIN)) ? xmlDUMB : xmlDISABLE;
    break;
  case ctRICHTEXT:
  case ctUNTEXT:
    return (m_options & (1<<M_OPTION_PLAIN)) ? xmlDISABLE : xmlDUMB;
    break;
  case ctHTML:
    return (m_options & (1<<M_OPTION_PLAIN)) ? xmlDISABLE : xmlHTML;
    break;
  case ctOTHER:
  case ctAPP:
  case ctXML:
  case ctSGML:
    return (m_options & (1<<M_OPTION_NOPLAIN)) ? xmlDUMB : xmlDISABLE;
    break;
  }
  return xmlUNDEF;
}

void reset_xml_character_filter(XML_Reset reset) {
  switch(reset) {
  case xmlRESET:
    xml.hide = VISIBLE;
    xml.state = TEXT;
    if( m_options & (1<<M_OPTION_HTML) ) {
      xml.parser = xpSMART;
    } else if( m_options & (1<<M_OPTION_XML) ) {
      xml.parser = xpDUMB;
    } else {
      xml.parser = xpSMART;
    }
    break;
  case xmlDISABLE:
    xml.state = DISABLED;
    break;
  case xmlSMART:
    if( (xml.state == DISABLED) ||
	((xml.parser != xpHTML) && (xml.parser != xpSMART)) ) {
      xml.state = TEXT;
      xml.hide = VISIBLE;
      xml.parser = xpSMART;
    }
    break;
  case xmlHTML:
    if( (xml.state == DISABLED) ||
	((xml.parser != xpHTML) && (xml.parser != xpSMART)) ) {
      xml.state = TEXT;
      xml.hide = VISIBLE;
      xml.parser = xpHTML;
    }
    break; 
  case xmlDUMB:
    if( (xml.state == DISABLED) ||
	(xml.parser != xpDUMB) ) {
      xml.state = TEXT;
      xml.hide = VISIBLE;
      xml.parser = xpDUMB;
      break;
    }
  case xmlUNDEF:
    /* ignore */
    break;
  }
}

void reset_mbox_line_filter() {
  mbox.state = msUNDEF;
  mbox.substate = msuUNDEF;
  mbox.header.type = mbox.body.type = ctUNDEF;
  mbox.header.encoding = mbox.body.encoding = ceUNDEF;
  mbox.prev_line_empty = 1;
  mbox.corruption_check = 0;
  mbox.skip_header = 0;
  mbox.skip_until_boundary = 0;
  mbox.strip_header_comment = 0;
  mbox.plainstate = psPLAIN;
  memset(&mbox.boundary, 0, sizeof(mbox.boundary));

  /* no need to reserve space for both char and wide char caches */
  if( m_options & (1<<M_OPTION_I18N) ) {
#if defined HAVE_MBRTOWC
    w_init_decoding_caches();
#endif
  } else {
    init_decoding_caches();
  }
}


/***********************************************************
 * MULTIBYTE FILE HANDLING FUNCTIONS                       *
 * this is suitable for any locale whose character set     *
 * encoding doesn't include NUL bytes inside characters    *
 ***********************************************************/



#define MULTIBYTE_EPSILON 10 /* enough for a multibyte char and a null char */

/* reads a text file as input and applies several filters. */
void process_file(FILE *input, 
		  int (*line_filter)(char *),
		  void (*character_filter)(char *), 
		  void (*word_fun)(char *, token_order_t, regex_count_t), 
		  char *(*pre_line_fun)(char *),
		  void (*post_line_fun)(char *)) {
  char *s, *pptextbuf;
  regex_count_t i;
  charbuf_len_t k;
  int eflag;
  token_order_t z, j, n, order;
  charbuf_len_t l;

  char *p, *q, *qq;
  char tok[MAX_TOKEN_LEN+2];

  regmatch_t pmatch[MAX_SUBMATCH];

  char ntok[MAX_TOKEN_LEN+2];
  char *nq;
  token_order_t nhow_many;

  long int e;
#if defined HAVE_LIBBOOST_REGEX
  charbuf_len_t wclen;
  wchar_t *wp;
  mbstate_t tok_shiftstate, input_shiftstate;
  charbuf_len_t tok_len;

  memset(&input_shiftstate, 0, sizeof(mbstate_t));
#endif

  e = 0;
  /* initialize the norex state */
  ntok[0] = DIAMOND;
  ntok[1] = '\0';
  nq = ntok + 1;
  nhow_many = 0;
 
  /* now start processing */
  while( !(cmd & (1<<CMD_QUITNOW)) && !feof(input) ) {

    process_pending_signal(input);

    /* read in a full line, allocating memory as necessary */
    textbuf[0] = '\0';
    s = textbuf;
    l = textbuf_len;
    k = 1;
    while( fgets(s, l, input) && (strlen(s) >= (l - 1)) ) {
      textbuf = realloc(textbuf, 2 * textbuf_len);
      if( !textbuf ) {
	fprintf(stderr, 
		"error: not enough memory for input line (%d bytes)\n",
		textbuf_len);
	exit(0);
      }

      s = textbuf + textbuf_len - (k++);
      l = textbuf_len;
      textbuf_len *= 2;

#if defined HAVE_MADVISE
      madvise(textbuf, sizeof(char) * textbuf_len, MADV_SEQUENTIAL);
#endif

    }

    /* preprocesses textbuf, optionally censors it */
    if( pre_line_fun ) {
      pptextbuf = (*pre_line_fun)(textbuf);
      if( !pptextbuf ) { continue; }
    } else {
      pptextbuf = textbuf;
    }

    /* next we check to see if this line should be skipped */
    if( *pptextbuf && (!line_filter || (*line_filter)(pptextbuf)) ) {
      /* now filter some of the characters in the current line */
      if( character_filter ) { (*character_filter)(pptextbuf); }

      if( (u_options & (1<<U_OPTION_DEBUG)) && (u_options & (1<<U_OPTION_CLASSIFY)) ) {
	fprintf(stdout, "%s", pptextbuf);
	if( pptextbuf[strlen(pptextbuf) - 1] != '\n' ) {
	  fprintf(stdout, "\n");
	}
      }

      /* repeat for each regular expression:
	 find all the instances of a matching substring */
#if defined HAVE_LIBBOOST_REGEX

      /* damn - we compiled wide character regexes, 
	 now we've got to use them (not really, but who in their right
	 mind would compile with boost if they're not going to convert
	 to wide characters anyway) */

      if( textbuf_len > wc_textbuf_len ) {
	wc_textbuf_len = textbuf_len;
	wc_textbuf = realloc(wc_textbuf, wc_textbuf_len * sizeof(wchar_t));
	if( !wc_textbuf ) {
	  fprintf(stderr, 
		  "error: not enough memory for wide character conversion "
		  "(%d bytes)\n",
		  wc_textbuf_len * sizeof(wchar_t));
	  exit(0);
	}

#if defined HAVE_MADVISE
	madvise(wx_textbuf, sizeof(wchar_t) * wc_textbuf_len, MADV_SEQUENTIAL);
#endif

      }

      /* convert as much as we can of the line into wide characters */
      s = pptextbuf;
      k = textbuf_len;
      wp = wc_textbuf;
      wclen = 0;
      /* since we ensured textbuf_len <= wctextbuf_len
	 there will never be overflow of wctextbuf below */
      while( k > 0 ) {
	l = mbrtowc(wp, s, k, &input_shiftstate);
	if( l > 0 ) {
	  wp++;
	  wclen++;
	  k -= l;
	  s += l;
	} else if( l == 0 ) {
	  break;
	} else if( l == -1 ) {
	  /* try to be robust */
	  s++; 
	  k--;
	  memset(&input_shiftstate, 0, sizeof(mbstate_t));
	} else if( l == -2) {
	  /* couldn't parse a complete character */
	  break;
	}
      }
      *wp = L'\0';

      /* now process the regex */
      for(i = 0; i < regex_count; i++) {
	k = 0;
	eflag = 0;
	/* see if a match */
	while( (k < wclen) && (regexec(&re[i].regex, wc_textbuf + k, 
				   MAX_SUBMATCH, pmatch, eflag) == 0) ) {
 	  /* all the submatches (delimited by brackets in the regex) 
 	     get converted, concatenated and the result gets word_fun'd */
	  q = tok;
	  *q++ = DIAMOND;
	  memset(&tok_shiftstate, 0, sizeof(mbstate_t));
 	  for(order = 0, z = 1; 
	      (z < MAX_SUBMATCH) && (pmatch[z].rm_so > -1); z++) {
	    if( !(re[i].submatches & (1<<z)) ) 
	      { continue; } else { order++; }
 	    /* transcribe the submatch into tok */
 	    for(j = pmatch[z].rm_so; j < pmatch[z].rm_eo; j++) {
 	      if( q < tok + MAX_TOKEN_LEN - MULTIBYTE_EPSILON ) {
		if( m_options & (1<<M_OPTION_CASEN) ) {
		  tok_len = wcrtomb(q, wc_textbuf[k + j], &tok_shiftstate); 
		} else {
		  tok_len = wcrtomb(q, towlower(wc_textbuf[k + j]), 
				    &tok_shiftstate); 
		}
		if( (tok_len > -1) ) {
		  q += tok_len;
		}
	      }
	    }
	    *q++ = DIAMOND;
	  }
 	  *q = '\0'; 

 	  /* now let each category process the token */
	  (*word_fun)(tok, order, i + 1); 

	  k += pmatch[0].rm_so + 1; /* advance string and repeat */
 	  eflag = REG_NOTBOL;
	}
      }
#else 

      /* the GNU regex routines expect an ordinary string */
      for(i = 0; i < regex_count; i++) {
	k = 0;
 	eflag = 0;
 	/* see if a match */
 	while( (k < l) && (regexec(&re[i].regex, pptextbuf + k, 
				   MAX_SUBMATCH, pmatch, eflag) == 0) ) {
 	  /* all the submatches (delimited by brackets in the regex) 
 	     get concatenated and the result gets word_fun'd */
 	  q = tok;
	  *q++ = DIAMOND;
 	  for(order = 0, z = 1; 
	      (z < MAX_SUBMATCH) && (pmatch[z].rm_so > -1); z++) {
	    if( !(re[i].submatches & (1<<z)) ) 
	      { continue; } else { order++; }
 	    /* transcribe the submatch into tok */
 	    for(j = pmatch[z].rm_so; j < pmatch[z].rm_eo; j++) {
 	      if( q < tok + MAX_TOKEN_LEN ) {
		if( m_options & (1<<M_OPTION_CASEN) ) {
		  *q++ = pptextbuf[k + j];
		} else {
		  *q++ = tolower(pptextbuf[k + j]);
		}
 	      }
 	    }
	    *q++ = DIAMOND;
 	  }
 	  *q = '\0'; 
 	  /* now let each category process the token */
	  (*word_fun)(tok, order, i + 1); 

	  k += pmatch[0].rm_so + 1; /* advance string and repeat */
 	  eflag = REG_NOTBOL;
 	}	      
      }
      
#endif

      /* default processing: reads tokens and passes them to
	 the word_fun */
      if( !(m_options & (1<<M_OPTION_USE_REGEX)) ) {
	p = pptextbuf;

	while( *p ) {
	  switch( good_char(p) ) {
	  case 1:
	    if( nq < ntok + MAX_TOKEN_LEN ) {
	      *nq++ = *p;
	    }
	    break;
	  case 2:
	    if( nq < ntok + MAX_TOKEN_LEN ) {
	      *nq++ = *p;
	    }
	    /* don't break, fall through to boundary case */
	  case 0:
	    if( *(nq - 1) != DIAMOND ) { /* token boundary */
	      *nq++ = DIAMOND;
	      *nq = '\0';
	      if( ngram_order == 1 ) {
		/* let each category process the token */
		(*word_fun)(ntok, (token_order_t)1, 0);
		nq = ntok;
		*nq++ = DIAMOND;
	      } else {
		if( ++nhow_many > ngram_order ) {
		  nhow_many--;
		  /* move all tokens down by one */
		  for(nq = ntok + 1; *nq != DIAMOND; nq++) {};
		  for(nq++, qq = ntok + 1; *nq; *qq++ = *nq++) {};
		  *qq = '\0';
		  nq = qq;
		}

		qq = ntok;
		for(n = nhow_many; n > 0; n--) {
		  /* let each category process the token */
		  (*word_fun)(qq, (token_order_t)n, 0);
		  qq++;
		  /* skip to next token and repeat */
		  while(*qq != DIAMOND ) { qq++; }
		}
	      }
	    }
	  }
	  p++;
	}
      }
      
    }

    /* now summarize this line if required */
    if( post_line_fun ) { (*post_line_fun)(pptextbuf); }

    if( cmd & (1<<CMD_RELOAD_CATS) ) {

      reload_all_categories();

      cmd &= ~(1<<CMD_RELOAD_CATS);
    }

  }

}


/***********************************************************
 * WIDE CHARACTER FILE HANDLING FUNCTIONS                  *
 * this is needed for any locale whose character set       *
 * encoding can include NUL bytes inside characters        *
 ***********************************************************/
#if defined HAVE_MBRTOWC


/* reads a text file as input, converting each line
into a wide character representation and applies several
filters. */
void w_process_file(FILE *input, 
		    int (*line_filter)(wchar_t *),
		    void (*character_filter)(wchar_t *), 
		    void (*word_fun)(char *, token_order_t, regex_count_t), 
		    char *(*pre_line_fun)(char *),
		    void (*post_line_fun)(char *)) {
  char *s, *pptextbuf;
  regex_count_t i;
  charbuf_len_t k;
  int eflag;
  token_order_t z, j, n, order;
  charbuf_len_t l;
  charbuf_len_t wclen;

  wchar_t *wp;
  mbstate_t tok_shiftstate, input_shiftstate;
  charbuf_len_t tok_len;
  char *q, *qq;
  char tok[MAX_TOKEN_LEN+2];

  regmatch_t pmatch[MAX_SUBMATCH];

  char *nq;
  char ntok[MAX_TOKEN_LEN+2];
  token_order_t nhow_many;

  /* initialize the norex state */
  ntok[0] = DIAMOND;
  ntok[1] = '\0';
  nq = ntok + 1;
  nhow_many = 0;

  memset(&input_shiftstate, 0, sizeof(mbstate_t));

  while( !feof(input) ) {

    /* read in a full line, allocating memory as necessary */
    textbuf[0] = '\0';
    s = textbuf;
    l = textbuf_len;
    k = 1;
    while( fgets(s, l, input) && (strlen(s) >= (l - 1)) ) {
      textbuf = realloc(textbuf, 2 * textbuf_len);
      if( !textbuf ) {
	fprintf(stderr, 
		"error: not enough memory for input line (%d bytes)\n",
		textbuf_len);
	exit(0);
      }

      s = textbuf + textbuf_len - (k++);
      l = textbuf_len;
      textbuf_len *= 2;

#if defined HAVE_MADVISE
      madvise(textbuf, sizeof(char) * textbuf_len, MADV_SEQUENTIAL);
#endif

    }

    /* preprocesses textbuf, optionally censors it */
    if( pre_line_fun ) {
      pptextbuf = (*pre_line_fun)(textbuf);
      if( !pptextbuf ) { continue; }
    } else {
      pptextbuf = textbuf;
    }

    /* now convert the line into a wide character string */
    if( textbuf_len > wc_textbuf_len ) {
      wc_textbuf_len = textbuf_len;
      wc_textbuf = realloc(wc_textbuf, wc_textbuf_len * sizeof(wchar_t));
      if( !wc_textbuf ) {
	fprintf(stderr, 
		"error: not enough memory for wide character conversion "
		"(%ld bytes)\n",
		(long int)(wc_textbuf_len * sizeof(wchar_t)));
	exit(0);
      }

#if defined HAVE_MADVISE
	madvise(wc_textbuf, sizeof(wchar_t) * wc_textbuf_len, MADV_SEQUENTIAL);
#endif

    }

    /* convert as much as we can of the line into wide characters */
    s = pptextbuf;
    k = textbuf_len;
    wp = wc_textbuf;
    wclen = 0;
    /* since we ensured textbuf_len <= wctextbuf_len
       there will never be overflow of wctextbuf below */
    while( k > 0 ) {
      l = mbrtowc(wp, s, k, &input_shiftstate);
      if( l > 0 ) {
	wp++;
	wclen++;
	k -= l;
	s += l;
      } else if( l == 0 ) {
	break;
      } else if( l == -1 ) {
	/* try to be robust */
	s++; 
	k--;
	memset(&input_shiftstate, 0, sizeof(mbstate_t));
      } else if( l == -2) {
	/* couldn't parse a complete character */
	break;
      }
    }
    *wp = L'\0';

    /* next we check to see if this line should be skipped */
    if( (wclen > 0) && 
	(!line_filter || (*line_filter)(wc_textbuf)) ) {
      /* now filter some of the characters in the current line */
      if( character_filter ) { (*character_filter)(wc_textbuf); }

      /* repeat for each regular expression:
	 find all the instances of a matching substring */
#if defined HAVE_LIBBOOST_REGEX

      for(i = 0; i < regex_count; i++) {
	k = 0;
	eflag = 0;
	/* see if a match */
	while( (k < wclen) && (regexec(&re[i].regex, wc_textbuf + k, 
				   MAX_SUBMATCH, pmatch, eflag) == 0) ) {
 	  /* all the submatches (delimited by brackets in the regex) 
 	     get converted, concatenated and the result gets word_fun'd */
	  q = tok;
	  *q++ = DIAMOND;
	  memset(&tok_shiftstate, 0, sizeof(mbstate_t));
 	  for(order = 0, z = 1; 
	      (z < MAX_SUBMATCH) && (pmatch[z].rm_so > -1); z++) {
	    if( !(re[i].submatches & (1<<z)) ) 
	      { continue; } else { order++; }
 	    /* transcribe the submatch into tok */
 	    for(j = pmatch[z].rm_so; j < pmatch[z].rm_eo; j++) {
 	      if( q < tok + MAX_TOKEN_LEN - MULTIBYTE_EPSILON ) {
		if( m_options & (1<<M_OPTION_CASEN) ) {
		  tok_len = wcrtomb(q, wc_textbuf[k + j], &tok_shiftstate); 
		} else {
		  tok_len = wcrtomb(q, towlower(wc_textbuf[k + j]), 
				    &tok_shiftstate); 
		}
		if( (tok_len > -1) ) {
		  q += tok_len;
		}
	      }
	    }
	    *q++ = DIAMOND;
	  }
 	  *q = '\0'; 

 	  /* now let each category process the token */
 	  (*word_fun)(tok, order, i + 1); 

	  k += pmatch[0].rm_so + 1; /* advance string and repeat */
 	  eflag = REG_NOTBOL;
	}
      }
#else 
      /* the GNU regex routines expect a multibyte string */
      if( textbuf_len > aux_textbuf_len ) {
	aux_textbuf_len = textbuf_len;
	aux_textbuf = realloc(aux_textbuf, aux_textbuf_len);
	if( !aux_textbuf ) {
	  fprintf(stderr, 
		  "error: not enough memory for auxiliary text buffer "
		  "(%d bytes)\n",
		  aux_textbuf_len);
	  exit(0);
	}
      }
      l = wcstombs(aux_textbuf, wc_textbuf, aux_textbuf_len - 1);
      aux_textbuf[aux_textbuf_len - 1] = '\0';

      for(i = 0; i < regex_count; i++) {
	k = 0;
 	eflag = 0;
 	/* see if a match */
 	while( (k < l) && (regexec(&re[i].regex, aux_textbuf + k, 
				   MAX_SUBMATCH, pmatch, eflag) == 0) ) {
 	  /* all the submatches (delimited by brackets in the regex) 
 	     get concatenated and the result gets word_fun'd */
 	  q = tok;
	  *q++ = DIAMOND;
 	  for(order = 0, z = 1; 
	      (z < MAX_SUBMATCH) && (pmatch[z].rm_so > -1); z++) {
	    if( !(re[i].submatches & (1<<z)) ) 
	      { continue; } else { order++; }
 	    /* transcribe the submatch into tok */
 	    for(j = pmatch[z].rm_so; j < pmatch[z].rm_eo; j++) {
 	      if( q < tok + MAX_TOKEN_LEN ) {
		if( m_options & (1<<M_OPTION_CASEN) ) {
		  *q++ = aux_textbuf[k + j];
		} else {
		  *q++ = tolower(aux_textbuf[k + j]);
		}
 	      }
 	    }
	    *q++ = DIAMOND; /* hope DIAMOND can never be matched inside regex */
 	  }
 	  *q = '\0'; 
 	  /* now let each category process the token */
	  (*word_fun)(tok, order, i + 1); 

	  k += pmatch[0].rm_so + 1; /* advance string and repeat */
 	  eflag = REG_NOTBOL;
 	}	      
      }
      
#endif

      /* default processing: reads tokens, converting them
	 to multibyte representation before passing them to
	 the word_fun */
      if( !(m_options & (1<<M_OPTION_USE_REGEX)) ) {
	wp = wc_textbuf;
	memset(&tok_shiftstate, 0, sizeof(mbstate_t));

	while( *wp ) {
	  switch( w_good_char(wp) ) {
	  case 1:
	    if( nq < ntok + MAX_TOKEN_LEN - MULTIBYTE_EPSILON ) {
	      tok_len = wcrtomb(nq, *wp, &tok_shiftstate); 
	      if( (tok_len > -1) ) {
		nq += tok_len;
	      }
	    }
	    break;
	  case 2:
	    if( nq < ntok + MAX_TOKEN_LEN - MULTIBYTE_EPSILON ) {
	      tok_len = wcrtomb(nq, *wp, &tok_shiftstate); 
	      if( (tok_len > -1) ) {
		nq += tok_len;
	      }
	    }
	    /* don't break, fall through to boundary case */
	  case 0:
	    if( *(nq - 1) != DIAMOND ) { /* token boundary */
	      *nq++ = DIAMOND;
	      *nq = '\0';
	      if( ngram_order == 1 ) {
		/* let each category process the token */
		(*word_fun)(ntok, (token_order_t)1, 0);
		nq = ntok;
		*nq++ = DIAMOND;
	      } else {
		if( ++nhow_many > ngram_order ) {
		  nhow_many--;
		  /* move all tokens down by one */
		  for(nq = ntok + 1; *nq != DIAMOND; nq++) {};
		  for(nq++, qq = ntok + 1; *nq; *qq++ = *nq++) {};
		  *qq = '\0';
		  nq = qq;
		}

		qq = ntok;
		for(n = nhow_many; n > 0; n--) {
		  /* let each category process the token */
		  (*word_fun)(qq, (token_order_t)n, 0);
		  qq++;
		  /* skip to next token and repeat */
		  while(*qq != DIAMOND ) { qq++; }
		}
	      }
	    }
	  }
	  wp++;
	}
      }

    }

    /* now summarize this line if required */
    if( post_line_fun ) { (*post_line_fun)(pptextbuf); }
      
  }
}

#endif /* DISABLE_WCHAR */
